from typing import Callable, Coroutine, Type

from django.db.models import ManyToOneRel
from django.db.models import Model as DjangoModel
from django.db.models.fields.related import RelatedField
from strawberry.types import Info
from strawberry_django.fields.field import StrawberryDjangoField

from .core.factory import BaseDjangoModelDataLoaderFactory
from .dataloaders import BasicPKDataLoader, BasicReverseFKDataLoader


class PKDataLoaderFactory(BaseDjangoModelDataLoaderFactory):
    """
    Base factory for simple PK relationship dataloaders. For example, get favourite fruit (Fruit model) of a User.

    EXAMPLE:
        CONSIDER DJANGO MODELS:
            class Fruit(models.Model):
                ...

            class User(models.Model):
                favourite_fruit = models.ForeignKey("Fruit", ...)

        THE FACTORY WOULD BE USED IN A FOLLOWING MANNER:
            @strawberry.django.type(models.User)
            class UserType:
                ...
                @strawberry.field
                async def favourite_fruit(self: "models.User", info: "Info") -> "FruitType":
                    loader = BasicPKDataLoaderFactory.get_loader_class('<django_app>.Fruit')
                    return await loader(context=info.context).load(self.favourite_fruit_id)
    """

    loader_class = BasicPKDataLoader

    @classmethod
    def get_loader_key(cls, model: Type["DjangoModel"], **kwargs):
        return model

    @classmethod
    def get_loader_class_kwargs(cls, model: Type["DjangoModel"], **kwargs):
        return {
            "model": model,
        }

    @classmethod
    def as_resolver(cls) -> Callable[["DjangoModel", Info], Coroutine]:
        async def resolver(root: "DjangoModel", info: "Info"):  # beware, first argument needs to be called 'root'
            field_data: "StrawberryDjangoField" = info._field
            relation: "RelatedField" = root._meta.get_field(field_name=field_data.django_name)
            pk = getattr(root, relation.attname)
            return await cls.get_loader_class(field_data.django_model)(context=info.context).load(pk)

        return resolver


class ReverseFKDataLoaderFactory(BaseDjangoModelDataLoaderFactory):
    """
    Base factory for reverse FK relationship dataloaders. For example, get blog posts of a User.

    EXAMPLE:
        CONSIDER DJANGO MODELS:
            class User(models.Model):
                ...

            class BlogPost(models.Model):
                published_by = models.ForeignKey("User", ...)

        THE FACTORY WOULD BE USED IN A FOLLOWING MANNER:
            @strawberry.django.type(models.User)
            class UserType:
                ...
                @strawberry.field
                async def blog_posts(self: "models.User", info: "Info") -> list["BlogPostType"]:
                    loader = BasicReverseFKDataLoaderFactory.get_loader_class(
                        '<app_name>.BlogPost',
                        reverse_path='published_by_id',
                    )
                    return await loader(context=info.context).load(self.pk)
    """

    loader_class = BasicReverseFKDataLoader

    @classmethod
    def get_loader_key(cls, model: Type["DjangoModel"], **kwargs):
        reverse_path = kwargs.get("reverse_path")
        if not reverse_path:
            raise ValueError(f"{cls.__name__}: 'reverse_path' not specified for reverse relation of {model.__name__}.")
        return f"{model}-{reverse_path}"

    @classmethod
    def get_loader_class_kwargs(cls, model: Type["DjangoModel"], **kwargs):
        return {
            "model": model,
            "reverse_path": kwargs["reverse_path"],
        }

    @classmethod
    def as_resolver(cls) -> Callable[["DjangoModel", Info], Coroutine]:
        async def resolver(root: "DjangoModel", info: "Info"):  # beware, first argument needs to be called 'root'
            field_data: "StrawberryDjangoField" = info._field
            relation: "ManyToOneRel" = root._meta.get_field(field_name=field_data.django_name)
            loader = cls.get_loader_class(field_data.django_model, reverse_path=relation.field.attname)
            return await loader(context=info.context).load(root.pk)

        return resolver



from typing import TYPE_CHECKING, Any, Callable, Coroutine, Hashable, Type, Union, cast

from django.apps import apps
from strawberry.types import Info

from strawberry_django_dataloaders.core.dataloader import BaseDataLoader

if TYPE_CHECKING:
    from django.db.models import Model as DjangoModel  # pragma: nocover


class BaseDataLoaderFactory:
    loader_class: Type["BaseDataLoader"]
    registered_dataloaders: dict[Hashable, Type["BaseDataLoader"]] = {}

    @classmethod
    def get_loader_class(cls, *args, **kwargs) -> Type["BaseDataLoader"]:
        """Generates dataloader classes at runtime when they are used the first time, later gets them from cache."""
        dataloader_key = cls.get_loader_key(*args, **kwargs)
        if dataloader_key not in cls.registered_dataloaders:
            loader_cls = cls._create_dataloader_cls(dataloader_key, *args, **kwargs)
            cls.registered_dataloaders[dataloader_key] = loader_cls
        return cls.registered_dataloaders[dataloader_key]

    @classmethod
    def _create_dataloader_cls(cls, dataloader_key: Hashable, *args, **kwargs) -> Type["BaseDataLoader"]:
        return cast(
            Type[BaseDataLoader],
            type(
                f"{dataloader_key}{cls.loader_class.__name__}",
                (cls.loader_class,),
                cls.get_loader_class_kwargs(*args, **kwargs),
            ),
        )

    @classmethod
    def get_loader_class_kwargs(cls, *args, **kwargs) -> dict:
        return kwargs

    @classmethod
    def get_loader_key(cls, *args, **kwargs) -> Hashable:
        """Return a unique key for the dataloader."""
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def as_resolver(cls, *args, **kwargs) -> Callable[[Any, Info], Coroutine]:
        """
        Return a dataloader as a callable to be used in the field definition as a resolver.

        Example:
            @strawberry.django.type(models.User)
            class UserType:
                favourite_fruit = strawberry.django.field(resolver=<Factory>.as_resolver(args, kwargs))
        """
        raise NotImplementedError  # pragma: nocover


class BaseDjangoModelDataLoaderFactory(BaseDataLoaderFactory):
    @classmethod
    def get_loader_class(
        cls,
        model: Union[Type["DjangoModel"], str],  # Django model directly, or 'app_name.DjangoModel'
        /,  # forces 'model' to be positional argument
        **kwargs,
    ):
        if isinstance(model, str):
            model = apps.get_model(model)
        return super().get_loader_class(model, **kwargs)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Type

from strawberry.django.context import StrawberryDjangoContext
from strawberry.django.views import AsyncGraphQLView

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse  # pragma: nocover

    from strawberry_django_dataloaders.core.dataloader import BaseDataLoader  # pragma: nocover


@dataclass
class DataloaderContext(StrawberryDjangoContext):
    dataloaders: dict[Type["BaseDataLoader"], "BaseDataLoader"] = field(default_factory=dict)


class DataloaderAsyncGraphQLView(AsyncGraphQLView):
    async def get_context(self, request: "HttpRequest", response: "HttpResponse") -> DataloaderContext:
        context: "StrawberryDjangoContext" = await super().get_context(request, response)
        return DataloaderContext(**context.__dict__)



from dataclasses import dataclass


class BaseError(Exception):
    message: str
    code: str


@dataclass
class UnsupportedRelationError(BaseError):
    message: str = "Unsupported relation"
    code: str = "unsupported_relation"



from collections import defaultdict

from asgiref.sync import sync_to_async
from django.db.models import Model as DjangoModel

from strawberry_django_dataloaders.core.dataloader import BaseDjangoModelDataLoader


class BasicPKDataLoader(BaseDjangoModelDataLoader):
    """
    Base loader for simple PK relationship (e.g. get TravianAlliance of TravianAccount).

    EXAMPLE - load Travian alliance of TravianAccount:
        1. DATALOADER DEFINITION
        class BasicPKTravianAllianceDataLoader(BasicPKDataLoader):
            model = TravianAlliance

        2. USAGE
        @strawberry.django.type(models.TravianAccount)
        class TravianAccountType:
            ...

            @strawberry.field
            async def travian_alliance(self: "models.TravianAccount", info: "Info") -> list["TravianVillageType"]:
                return await BasicPKTravianAllianceDataLoader(context=info.context).load(self.travian_alliance_id)

    """

    @classmethod
    @sync_to_async
    def load_fn(cls, keys: list[str]) -> list[DjangoModel | None]:
        instances: list["DjangoModel"] = list(cls.model.objects.filter(pk__in=keys))
        # ensure instances are ordered in the same way as input 'keys'
        id_to_instance: dict[str, "DjangoModel"] = {inst.pk: inst for inst in instances}
        return [id_to_instance.get(id_) for id_ in keys]


class BasicReverseFKDataLoader(BaseDjangoModelDataLoader):
    """
    Base loader for reversed FK relationship (e.g. get BlogPosts of a User).

    EXAMPLE - load blog posts of account:
        1. DATALOADER DEFINITION
        class BlogPostsBasicReverseFKDataLoader(BasicReverseFKDataLoader):
            model = BlogPost
            reverse_path = 'user_id'

        2. USAGE
        @strawberry.django.type(models.User)
        class UserType:
            ...

            @strawberry.field
            async def blog_posts(self: "models.User", info: "Info") -> list["BlogPostType"]:
                return await BlogPostsBasicReverseFKDataLoader(context=info.context).load(self.pk)
    """

    reverse_path: str  # path to the 'parent' model from the reverse relationship

    @classmethod
    @sync_to_async
    def load_fn(cls, keys: list[str]) -> list[list[DjangoModel]]:
        instances: list["DjangoModel"] = list(cls.model.objects.filter(**{f"{cls.reverse_path}__in": keys}))
        # ensure that instances are ordered the same way as input 'ids'
        id_to_instances: dict[str, list["DjangoModel"]] = defaultdict(list)
        for instance in instances:
            id_to_instances[getattr(instance, cls.reverse_path)].append(instance)
        return [id_to_instances.get(key, []) for key in keys]



from typing import TYPE_CHECKING, Any

import strawberry.django
from strawberry import UNSET

from . import exceptions, factories

if TYPE_CHECKING:
    from django.db.models import Model as DjangoModel  # pragma: nocover
    from django.db.models.fields.related import RelatedField  # pragma: nocover
    from strawberry.types import Info  # pragma: nocover
    from strawberry_django.fields.field import StrawberryDjangoField  # pragma: nocover


async def get_dataloader_resolver(root: "DjangoModel", info: "Info"):
    field_data: "StrawberryDjangoField" = info._field
    relation: "RelatedField" = root._meta.get_field(field_name=field_data.django_name)
    if relation.many_to_one or relation.one_to_one:
        return await factories.PKDataLoaderFactory.as_resolver()(root, info)
    elif relation.one_to_many:
        return await factories.ReverseFKDataLoaderFactory.as_resolver()(root, info)
    else:
        raise exceptions.UnsupportedRelationError(f"Unsupported relation on {relation.__repr__()}.")


def auto_dataloader_field(
    resolver=get_dataloader_resolver,
    *,
    name=None,
    field_name=None,
    filters=UNSET,
    default=UNSET,
    **kwargs,
) -> Any:
    """
    A field which has automatic dataloader resolver based on the relationship type (one-to-one, one-to-many, etc.).

    EXAMPLE:
        CONSIDER DJANGO MODELS:
            class Fruit(BaseTestModel):
                plant = models.OneToOneField("FruitPlant", ...)
                color = models.ForeignKey("Color", related_name="fruits", ...)
                varieties = models.ManyToManyField("FruitVariety", related_name="fruits")

            class FruitEater(BaseTestModel):
                favourite_fruit = models.ForeignKey("Fruit", related_name="eaters", ...)

        DEFINE THE STRAWBERRY TYPE AS:
            @strawberry_django.type(models.Fruit)
            class FruitTypeAutoDataLoaderFields:
                plant: FruitPlantType = fields.auto_dataloader_field()
                color: ColorType = fields.auto_dataloader_field()
                varieties: list[FruitVarietyType] = fields.auto_dataloader_field()
                eaters: list[FruitEaterType] = fields.auto_dataloader_field()
    """
    return strawberry.django.field(
        resolver=resolver,
        name=name,
        field_name=field_name,
        filters=filters,
        default=default,
        **kwargs,
    )



from typing import TYPE_CHECKING, Type

from asgiref.sync import sync_to_async
from strawberry.dataloader import DataLoader

if TYPE_CHECKING:
    from django.db.models import Model as DjangoModel  # pragma: nocover

    from strawberry_django_dataloaders.views import DataloaderContext  # pragma: nocover


class BaseDataLoader(DataLoader):
    _instance_cache = None

    def __new__(cls, context: "DataloaderContext", force_new: bool = False) -> "BaseDataLoader":
        """
        Returns a dataloader instance.
        Takes the instance from request context cache or creates a new one if it does not exist there yet.
        This makes the dataloader "semi-singleton" in the sense that they are singleton in the context of each request.
        """
        if force_new or cls not in context.dataloaders:
            context.dataloaders[cls] = super().__new__(cls)
        return context.dataloaders[cls]

    def __init__(self, context: "DataloaderContext", **kwargs):
        if self._instance_cache is None:
            self._instance_cache = context.dataloaders[self.__class__]
        super().__init__(**kwargs)


class BaseDjangoModelDataLoader(BaseDataLoader):
    model: Type["DjangoModel"] = NotImplemented

    def __init__(self, *args, **kwargs):
        super().__init__(*args, load_fn=self.load_fn, **kwargs)

    @classmethod
    @sync_to_async
    def load_fn(cls, keys: list[str]):
        raise NotImplementedError  # pragma: nocover
